module Main exposing (..)

import Browser
import Data.DataLoader exposing (decode)
import Html exposing (Html, a, button, div, i, nav, text)
import Html.Attributes
import Html.Events exposing (onClick)
import Http
import Model exposing (Global, Local, Model, Progress(..), View(..))
import Msg exposing (Msg(..))
import Triple
import Utils exposing (buildErrorMessage)
import Views.BarPlot as BarPlotView
import Views.DataTableView as DataTableView
import Views.ForceDirectedGraph as ForceDirectedGraphView
import Views.LinePlot as LinePlotView
import Views.MapView as MapView



-- MAIN


main : Program () Model Msg
main =
    Browser.element
        { init = init
        , update = update
        , view = view
        , subscriptions = subscriptions
        }



-- MODEL


init : () -> ( Model, Cmd Msg )
init () =
    let
        global : Global
        global =
            { view = Home
            , progress = Loading
            , data = []
            , selectedCountry = Nothing
            , selectedConsulateCountry = Nothing
            }

        local : Local
        local =
            { map = MapView.init |> Tuple.first
            , dataTable = DataTableView.init
            , forcedirectedgraph = ForceDirectedGraphView.firstInit
            , barplot = BarPlotView.init
            , lineplot = LinePlotView.init
            }

        subCmd : List (Cmd Msg)
        subCmd =
            [ MapView.init ] |> List.map Tuple.second

        cmd : Cmd Msg
        cmd =
            List.append
                [ Http.get
                    { url = "/schengen_data.csv"
                    , expect = Http.expectString DataReceived
                    }
                ]
                subCmd
                |> Cmd.batch

        model : Model
        model =
            ( global, local )
    in
    ( model, cmd )



-- UPDATE


update : Msg -> Model -> ( Model, Cmd Msg )
update msg ( global, local ) =
    case msg of
        ViewChange newView ->
            case newView of
                ForceDirectedGraph ->
                    let
                        fdg =
                            local.forcedirectedgraph

                        loadingfdg =
                            { fdg | loading = True }
                    in
                    ( ( { global | view = newView }, { local | forcedirectedgraph = loadingfdg } )
                    , ForceDirectedGraphView.calculateModel global.data loadingfdg |> Cmd.map ForceDirectedGraphView.propagate
                    )

                _ ->
                    ( ( { global | view = newView }, local ), Cmd.none )

        DataReceived (Ok rawData) ->
            case decode rawData of
                -- Pull the map data ONLY AFTER the Visa data was received.
                Ok data ->
                    ( ( { global | progress = Success, data = data }
                      , { local | forcedirectedgraph = ForceDirectedGraphView.firstInit }
                      )
                    , MapView.getMapData |> Cmd.map GotMap
                    )

                Err _ ->
                    ( ( { global | progress = Failure "Failed to parse data." }, local ), Cmd.none )

        DataReceived (Err error) ->
            ( ( { global | progress = buildErrorMessage error |> Failure }, local ), Cmd.none )

        GotMap subMsg ->
            let
                updateResult =
                    MapView.update subMsg global local.map
            in
            ( ( Triple.first updateResult, { local | map = Triple.second updateResult } ), Triple.third updateResult )

        GotDataTable subMsg ->
            let
                updateResult =
                    DataTableView.update subMsg global local.dataTable
            in
            ( ( Triple.first updateResult, { local | dataTable = Triple.second updateResult } ), Triple.third updateResult )

        GotForceDirectedGraph subMsg ->
            let
                updateResult =
                    ForceDirectedGraphView.update subMsg global local.forcedirectedgraph
            in
            ( ( Triple.first updateResult, { local | forcedirectedgraph = Triple.second updateResult } ), Triple.third updateResult )

        GotBarPlot subMsg ->
            let
                updateResult =
                    BarPlotView.update subMsg global local.barplot
            in
            ( ( Triple.first updateResult, { local | barplot = Triple.second updateResult } ), Triple.third updateResult )

        GotLinePlot subMsg ->
            let
                updateResult =
                    LinePlotView.update subMsg global local.lineplot
            in
            ( ( Triple.first updateResult, { local | lineplot = Triple.second updateResult } ), Triple.third updateResult )



-- VIEW


viewNav : Model -> Html Msg
viewNav ( global, _ ) =
    let
        activeClass : View -> String
        activeClass v =
            if global.view == v then
                "navbar-active"

            else
                "navbar-inactive"
    in
    nav [ Html.Attributes.class "navbar" ]
        [ button [ activeClass Home |> Html.Attributes.class, ViewChange Home |> onClick ]
            [ i [ Html.Attributes.class "fa-solid fa-house fa-xl" ] [] ]
        , div [ Html.Attributes.class "navbar-title" ]
            [ text "Schengen Visa Statistics" ]
        , div [ Html.Attributes.class "navbar-view-icons" ]
            [ button [ activeClass MapView |> Html.Attributes.class, ViewChange MapView |> onClick ]
                [ i [ Html.Attributes.class "fa-solid fa-map fa-xl" ] [] ]
            , button [ activeClass ForceDirectedGraph |> Html.Attributes.class, ViewChange ForceDirectedGraph |> onClick ]
                [ i [ Html.Attributes.class "fa-solid fa-hexagon-nodes fa-xl" ] [] ]
            , button [ activeClass BarPlot |> Html.Attributes.class, ViewChange BarPlot |> onClick ]
                [ i [ Html.Attributes.class "fa-solid fa-chart-simple fa-xl" ] [] ]
            , button [ activeClass LinePlot |> Html.Attributes.class, ViewChange LinePlot |> onClick ]
                [ i [ Html.Attributes.class "fa-solid fa-chart-line fa-xl" ] [] ]
            , button [ activeClass DataTable |> Html.Attributes.class, ViewChange DataTable |> onClick ]
                [ i [ Html.Attributes.class "fa-solid fa-table fa-xl" ] [] ]
            ]
        ]


viewContent : Model -> Html Msg
viewContent ( global, local ) =
    let
        message =
            Html.div
                [ Html.Attributes.style "margin" "auto" ]
                (case global.progress of
                    Loading ->
                        [ Html.div [ Html.Attributes.class "loader" ] []
                        , Html.div
                            [ Html.Attributes.style "position" "relative"
                            , Html.Attributes.style "left" "1.5rem"
                            , Html.Attributes.style "bottom" "5rem"
                            ]
                            [ Html.text "Loading Data..." ]
                        ]

                    Failure msg ->
                        [ Html.text (msg |> (++) "Error while loading data. ") ]

                    Success ->
                        [ Html.text "Data was loaded successfully!" ]
                )
    in
    case global.view of
        Home ->
            div [ Html.Attributes.class "content" ]
                [ message ]

        MapView ->
            div [ Html.Attributes.class "content" ]
                [ MapView.view global local.map ]

        DataTable ->
            div [ Html.Attributes.class "content" ]
                [ DataTableView.view global local.dataTable ]

        ForceDirectedGraph ->
            div [ Html.Attributes.class "content" ]
                [ ForceDirectedGraphView.view global local.forcedirectedgraph ]

        BarPlot ->
            div [ Html.Attributes.class "content" ]
                [ BarPlotView.view global local.barplot ]

        LinePlot ->
            div [ Html.Attributes.class "content" ]
                [ LinePlotView.view global local.lineplot ]


viewFooter : Model -> Html Msg
viewFooter model =
    div [ Html.Attributes.class "footer noselect" ]
        [ text "Erstellt von Aamon Hoffmann, Damian Haj-Houssin und Florian Raab mit "
        , a
            [ Html.Attributes.target "_"
            , Html.Attributes.href "https://elm-lang.org/docs"
            ]
            [ text "Elm" ]
        , text " und gebundlet mit "
        , a
            [ Html.Attributes.target "_"
            , Html.Attributes.href "https://parceljs.org/languages/elm/"
            ]
            [ text "Parcel" ]
        , text "."
        ]


view : Model -> Html Msg
view model =
    div [ Html.Attributes.class "wrapper" ]
        [ viewNav model
        , viewContent model
        , viewFooter model
        ]



-- SUBSCRIPTIONS


subscriptions : Model -> Sub Msg
subscriptions ( global, local ) =
    case global.view of
        DataTable ->
            DataTableView.subscriptions global local.dataTable

        ForceDirectedGraph ->
            ForceDirectedGraphView.subscriptions global local.forcedirectedgraph

        MapView ->
            MapView.subscriptions global local.map

        _ ->
            Sub.none
