module Views.MapView exposing (view, update, init, subscriptions, getMapData)

import Http
import Html exposing (Html, div, text, select, option, span)
import Html.Lazy
import Html.Events exposing (onInput)
import Html.Attributes
import Color exposing (Color)
import Config.MapViewConfig as Local exposing (Model, MapFrame, MapCountry, CountryType(..), Progress(..), ChernoffConfig)
import Curve
import Dict exposing (Dict)
import Data.Query
import Data.DataLoader exposing (Entry)
import Dict.Extra
import List.Extra
import Model exposing (Global, View(..))
import Msg exposing (Msg(..))
import SubPath exposing (SubPath)
import TypedSvg exposing (svg, g, circle, polygon)
import TypedSvg.Core exposing (Svg)
import TypedSvg.Types exposing (Paint(..), Transform(..), Opacity(..))
import TypedSvg.Attributes exposing (fill, class, points, viewBox, transform, stroke, opacity)
import TypedSvg.Attributes.InPx exposing (r, strokeWidth)
import MapLoader exposing (Country, Polygon, decoder, synonym, toCountries)
import Scale exposing (ContinuousScale)
import Statistics
import Utils exposing (buildErrorMessage, mpr, flag, codeToName)
import Svg
import Svg.Events exposing (onMouseOver, onMouseOut, onClick)
import Zoom
import Task

w : Float
w = 1200

h : Float
h = 800

wideExtent : List Float -> ( Float, Float )
wideExtent values =
    let
        extent : ( Float, Float )
        extent =
            Maybe.withDefault ( 0, 0 ) (Statistics.extent values)

        minVal : Float
        minVal =
            Tuple.first extent

        maxVal : Float
        maxVal =
            Tuple.second extent
    in
    ( minVal, maxVal )

propagate : Local.Msg -> Msg
propagate msg =
    GotMap msg

getMapData : Cmd Local.Msg
getMapData =
    Http.get
        { url = "/countries.geojson"
        , expect = Http.expectJson Local.MapDataReceived decoder
        }

init : ( Model, Cmd Msg )
init =
    let
        scale : Float
        scale = 4.5

        translate : { x: Float, y: Float }
        translate =
            -- Guesswork...
            { x = -3/7 * w * scale, y = -1/16 * h * scale }

        state : Model
        state =
            { countries = []
            , progress = Loading
            , hoveredCountry = Nothing
            , selectedYear = Nothing
            , selectedConsulate = Nothing
            , zoom = Zoom.init { width = w, height = h }
                |> Zoom.scaleExtent 1 100
                |> Zoom.translateExtent ( ( 0, 0 ), ( w, h ) )
                |> Zoom.setTransform Zoom.instantly { scale = scale, translate = translate }
            }
    in
    ( state, Cmd.none )


merge : List Entry -> List Country -> List MapCountry
merge data countries =
    let
        schengen : Dict String (List Entry)
        schengen = Data.Query.groupBySchengen data

        consulates : Dict String (List Entry)
        consulates = Data.Query.groupByConsulateCountry data

        convert : Country -> MapCountry
        convert country =
            case Dict.Extra.find (\c _ -> synonym country c) schengen of
                Just ( _, entries ) -> MapCountry country SchengenState entries
                Nothing ->
                    case Dict.Extra.find (\c _ -> synonym country c) consulates of
                        Just ( _, entries ) -> MapCountry country ConsulateCountry entries
                        Nothing -> MapCountry country Unknown []
    in List.map convert countries

update : Local.Msg -> Global -> Model -> ( Global, Model, Cmd Msg )
update msg global model = 
    case msg of
        Local.NoOp ->
            (global, model, Cmd.none)

        Local.MapDataReceived (Ok data) ->
            (global, { model | progress = Success, countries = data |> toCountries |> merge global.data }, Cmd.none)

        Local.MapDataReceived (Err err) ->
            (global, { model | progress = buildErrorMessage err |> Failure }, Cmd.none)

        Local.YearChange year ->
            (global, { model | selectedYear = year }, Cmd.none)

        Local.ConsulateChange consulate ->
            let updatedEntry = model.selectedConsulate |> Maybe.map (Tuple.mapSecond (\_ -> consulate)) in
            (global, { model | selectedConsulate = updatedEntry }, Cmd.none)
        
        Local.HoverCountry country ->
            (global, { model | hoveredCountry = country }, Cmd.none)

        Local.ClickCountry country ->
            case country.variant of
                -- Switch view to LinePlot.
                SchengenState ->
                    let
                        task : Cmd Msg
                        task =
                            Task.perform (ViewChange LinePlot |> always) (Task.succeed ())
                    in 
                    ({ global | selectedCountry = Just country.context.meta.iso_a2 }, model, task)
                -- Set selected consulate country. Deselect if the same country was clicked again.
                ConsulateCountry ->
                    if Just country == (model.selectedConsulate |> Maybe.map Tuple.first)
                    then ( { global | selectedConsulateCountry = Nothing }, { model | selectedConsulate = Nothing }, Cmd.none) -- Deselect.
                    else ( { global | selectedConsulateCountry = Just country.context.meta.iso_a2 }, { model | selectedConsulate = Just ( country, Nothing ) }, Cmd.none) -- Select.
                -- Do nothing.
                _ -> (global, model, Cmd.none)

        Local.MapZoom zoom ->
            (global, { model | zoom = Zoom.update zoom model.zoom }, Cmd.none )

xScale : List Float -> ContinuousScale Float
xScale values =
    Scale.linear ( 0, w ) (wideExtent values)

yScale : List Float -> ContinuousScale Float
yScale values =
    Scale.linear ( h, 0 ) (wideExtent values)

computeFrame : List Country -> MapFrame
computeFrame countries =
    let
        coords = countries
            |> List.map .polygons
            |> List.concat -- Concatenate Lists of Polygons.
            |> List.concat -- Concatenate Lists of Polygon Point Coordinates.

        xScaleLocal : ContinuousScale Float
        xScaleLocal =
            coords |> List.map Tuple.first |> xScale
            
        yScaleLocal : ContinuousScale Float
        yScaleLocal =
            coords |> List.map Tuple.second |> yScale
    in
    MapFrame xScaleLocal yScaleLocal

fitPoint : MapFrame -> ( Float, Float ) -> ( Float, Float )
fitPoint frame point =
    point |> Tuple.mapBoth (Scale.convert frame.xScale) (Scale.convert frame.yScale)

fitPolygon : MapFrame -> Polygon -> Polygon
fitPolygon frame polygon =
    polygon |> List.map (fitPoint frame)

viewChernoff : Float -> ChernoffConfig -> Svg Local.Msg
viewChernoff scale config =
    let
        mouth : SubPath
        mouth =
            let
                height = 1/8 + (7/8 - 1/8) * config.mouth.angle
                width = 1/2 * config.mouth.length
            in
            Curve.quadraticBezier ( -width, 1/2 ) [ ( ( 0, height ), ( width, 1/2 ) ) ]

        nose : SubPath
        nose =
            let left = -1/4 + (1/4 - (-1/4)) * config.nose in
            Curve.linearClosed [ ( 0, 0 ), ( left, 1/8 ), ( 0, 1/4 ) ]

        eyeRadius : Float -> Float
        eyeRadius size =
            1/8 + (1/4 - 1/8) * size

        pupil : Svg Local.Msg
        pupil =
            circle
                [ r <| 0.01
                , fill <| PaintNone
                , stroke <| Paint Color.black
                , strokeWidth 0.05
                ] []

        eyeLeft : Svg Local.Msg
        eyeLeft =
            let radius = eyeRadius config.eyes.left in
            g [ class [ "eye-left" ], transform [ Translate (-1/3) (-1/3) ] ]
                [ pupil
                , circle
                    [ r <| radius
                    , fill <| PaintNone
                    , stroke <| Paint Color.black
                    , strokeWidth 0.1
                    ] []
                ]

        eyeRight : Svg Local.Msg
        eyeRight =
            let radius = eyeRadius config.eyes.right in
            g [ class [ "eye-right" ], transform [ Translate (1/3) (-1/3) ] ]
                [ pupil
                , circle
                    [ r <| radius
                    , fill <| PaintNone
                    , stroke <| Paint Color.black
                    , strokeWidth 0.1
                    ] []
                ]
    in
    g [ class [ "chernoff" ], transform [ Translate config.x config.y, Scale scale scale ] ]
        [ circle
            [ r <| 1
            , fill <| PaintNone
            , stroke <| Paint Color.black
            , strokeWidth 0.1
            , class [ "face-outline" ]
            ] []
        , g [ class [ "face-eyes" ] ]
            [ eyeLeft, eyeRight ]
        , SubPath.element mouth
            [ fill <| PaintNone
            , stroke <| Paint Color.black
            , strokeWidth 0.1
            , class [ "face-mouth" ]
            ]
        , SubPath.element nose
            [ fill <| PaintNone
            , stroke <| Paint Color.black
            , strokeWidth 0.07
            , class [ "face-nose" ]
            ]
        ]

viewPolygonLayer : Local.Model -> MapFrame -> List MapCountry -> List ( Svg Local.Msg )
viewPolygonLayer model frame countries =
    let
        drawCountry : MapCountry -> Svg Local.Msg
        drawCountry country =
            let
                handleClick : Local.Msg
                handleClick =
                    Local.ClickCountry country

                handleMouseOver : Local.Msg
                handleMouseOver =
                    country.context |> Just >> Local.HoverCountry

                handleMouseLeave : Local.Msg
                handleMouseLeave =
                    Local.HoverCountry Nothing

                color : Color
                color =
                    case country.variant of
                        SchengenState -> Color.darkGreen
                        ConsulateCountry -> Color.purple
                        Unknown -> Color.gray

                strokeAttrs : { color: Color, width: Float }
                strokeAttrs =
                    let
                        isConsulateSelected : Bool
                        isConsulateSelected = 
                            case model.selectedConsulate of
                                Nothing -> False
                                Just consulateEntry ->
                                    Tuple.first consulateEntry == country
                    in
                    if isConsulateSelected
                    then { color = Color.lightBlue, width = 1 }
                    else { color = Color.white, width = 0.01 }

                polygons : List Polygon
                polygons =
                    country.context.polygons |> List.map (fitPolygon frame)

                drawPolygon : Polygon -> Svg Local.Msg
                drawPolygon poly =
                    let
                        opac : Float
                        opac =
                            if model.hoveredCountry == Just country.context
                            then 0.55 else 1
                    in
                    polygon
                        [ class [ "polygon" ]
                        , poly |> points
                        , onClick handleClick
                        , onMouseOver handleMouseOver
                        , onMouseOut handleMouseLeave
                        , fill <| Paint color
                        , opacity <| Opacity <| opac
                        , strokeWidth <| strokeAttrs.width
                        , stroke <| Paint <| strokeAttrs.color
                        ] []
            in
            g [ class [ "country" ] ]
                (List.map drawPolygon polygons)
    in
    List.map drawCountry countries

viewChernoffLayer : Local.Model -> MapFrame -> List MapCountry -> List ( Svg Local.Msg )
viewChernoffLayer model frame preCountries =
    let        
        applyFilter : MapCountry -> MapCountry
        applyFilter country =
            let
                keep : Entry -> Bool
                keep entry =
                    let
                        fitsConsulate : Bool
                        fitsConsulate =
                            case model.selectedConsulate of
                                Just ( consulateCountry, Nothing ) ->
                                    synonym consulateCountry.context entry.consulateCountry
                                Just ( consulateCountry, Just consulate ) ->
                                    synonym consulateCountry.context entry.consulateCountry && entry.consulate == consulate
                                _ -> True
                        fitsYear : Bool
                        fitsYear =
                            case model.selectedYear of
                                Nothing -> True
                                Just year -> entry.year == year
                    in
                    List.all identity [ fitsConsulate, fitsYear ]
            in
            { country | entries = country.entries |> List.filter keep }

        countries : List MapCountry
        countries =
            List.map applyFilter preCountries


        maxAppliedSum : Float
        maxAppliedSum =
            countries
                |> List.map .entries
                |> List.map Data.Query.sumTotalVisasApplied
                |> List.maximum
                |> Maybe.withDefault 0

        maxIssuedStd : Float
        maxIssuedStd =
            countries
                |> List.map .entries
                |> List.map Data.Query.stdTotalVisasIssued 
                |> List.maximum
                |> Maybe.withDefault 0

        drawChernoff : MapCountry -> Svg Local.Msg
        drawChernoff country =
            let
                polygons : List Polygon
                polygons =
                    country.context.polygons |> List.map (fitPolygon frame)

                handleMouseOver : Local.Msg
                handleMouseOver =
                    country.context |> Just >> Local.HoverCountry

                handleMouseLeave : Local.Msg
                handleMouseLeave =
                    Local.HoverCountry Nothing

                ( cx, cy ) = ( country.context.meta.labelX, country.context.meta.labelY )
                    |> fitPoint frame

                eye : Float
                eye =
                    -- Reliability of issued rate for country w.r.t.
                    -- different years, consulates, ...
                    1 - (Data.Query.stdTotalVisasIssued country.entries
                            / maxIssuedStd)

                config : ChernoffConfig
                config =
                    { x = cx
                    , y = cy
                    , mouth =
                        { angle = Data.Query.sumTotalVisasIssued country.entries
                            / Data.Query.sumTotalVisasApplied country.entries
                        , length = Data.Query.sumTotalVisasApplied country.entries
                            / maxAppliedSum
                        }
                    , nose = Data.Query.sumTotalIssuedMEV country.entries
                        / Data.Query.sumTotalVisasIssued country.entries
                    , eyes =
                        { left = eye, right = eye }
                    }

                scale : Float
                scale = mpr (polygons |> List.concat) ( cx, cy ) / 1.5

                opac : Float
                opac =
                    if model.hoveredCountry == Just country.context
                    then 1 else 0.3
            in
            g
                [ class [ "chernoff" ]
                , opacity <| Opacity <| opac
                , onMouseOut <| handleMouseLeave
                , onMouseOver <| handleMouseOver
                ]
                [ viewChernoff scale config ]
    in
    List.map drawChernoff countries

viewMap : Global -> Model -> Svg Local.Msg
viewMap global model =
    let
        frame : MapFrame
        frame = model.countries |> List.map .context |> computeFrame

        polygonLayer =
            viewPolygonLayer model frame model.countries
        chernoffLayer = model.countries
            |> List.filter (.variant >> (==) SchengenState)
            |> viewChernoffLayer model frame
    in
    svg (viewBox 0 0 w h :: class [ "globe" ] :: Zoom.events model.zoom Local.MapZoom)
        [ g [ Zoom.transform model.zoom ]
            (List.concat [ polygonLayer, chernoffLayer ])
        ]
        

viewHoverMsg : Model -> Html Msg
viewHoverMsg model =
    case model.hoveredCountry of
        Nothing ->
            text "Momentan nichts fokussiert."
        
        Just country ->
            "Maus befindet sich über: "
                ++ codeToName country.meta.iso_a2
                ++ " "
                ++ flag country.meta.iso_a2
                |> text

viewYearSelect : List MapCountry -> Maybe Int -> Html Msg
viewYearSelect countries selectedYear =
    let
        years : List Int
        years =
            countries
                |> List.map (.entries >> List.map .year)
                |> List.concat
                |> List.Extra.unique
                |> List.sort
                |> List.reverse

        default : String
        default = 
            "Alle Jahre"

        handleYearChange : String -> Local.Msg
        handleYearChange rawYear =
            if rawYear == default
            then Local.YearChange Nothing
            else case String.toInt rawYear of
                Nothing -> Local.NoOp
                Just year -> year |> Just |> Local.YearChange

        viewOption : Maybe Int -> Html Msg
        viewOption yearOption =
            let
                isSelected : Bool
                isSelected =
                    yearOption == selectedYear

                value : String
                value =
                    yearOption
                    |> Maybe.map String.fromInt
                    |> Maybe.withDefault default
            in
            option
                [ Html.Attributes.value value
                , Html.Attributes.selected isSelected
                ] [ text value ]
    in
    select [ onInput (handleYearChange >> propagate) ]
        (viewOption Nothing :: List.map (Just >> viewOption) years)


viewYearFilter : Model -> Html Msg
viewYearFilter model =
    div [ Html.Attributes.class "filter-year" ]
        [ text "Jahr: "
        , Html.Lazy.lazy2
            viewYearSelect model.countries model.selectedYear
        ]


viewConsulateSelect : MapCountry -> Maybe String -> Html Msg
viewConsulateSelect country selectedConsulate =
    let
        default : String
        default = "Alle Konsulate"

        consulates : List String
        consulates =
            country.entries
                |> List.map .consulate
                |> List.Extra.unique
                |> List.sort 

        handleConsulateChange : String -> Local.Msg
        handleConsulateChange consulate =
            if consulate == default
            then Local.ConsulateChange Nothing
            else consulate |> Just |> Local.ConsulateChange

        viewOption : Maybe String -> Html Msg
        viewOption consulateOption =
            let
                isSelected : Bool
                isSelected =
                    consulateOption == selectedConsulate

                value : String
                value =
                    Maybe.withDefault default consulateOption
            in
            option
                [ Html.Attributes.value value
                , Html.Attributes.selected isSelected
                ] [ text value ]
    in
    select [ onInput (handleConsulateChange >> propagate) ]
        (viewOption Nothing :: List.map (Just >> viewOption) consulates)


viewConsulateFilter : Model -> Html Msg
viewConsulateFilter model =
    div [ Html.Attributes.class "filter-consulate" ]
        ( case model.selectedConsulate of
            Nothing -> [ text "Kein Konsulatsland ausgewählt." ]
            Just ( country, consulate ) ->
                [ text "Konsulate: "
                , viewConsulateSelect country consulate
                ]
        )


view : Global -> Model -> Html Msg
view global model =
    case model.progress of
        Loading -> div [] [ text "Loading map data..." ]
        Failure str -> div [] [ text str ]
        Success ->
            div [ Html.Attributes.class "map" ]
                [ span [] [ viewHoverMsg model ]
                , div [ Html.Attributes.class "filter" ]
                    [ viewYearFilter model
                    , viewConsulateFilter model
                    ]
                , viewMap global model |> Svg.map propagate
                ]

subscriptions : Global -> Model -> Sub Msg
subscriptions global model =
    Zoom.subscriptions model.zoom Local.MapZoom |> Sub.map propagate